# MemoryCache.py,v 1.3 2003/06/01 10:29:39 jamie v0.9
__version__ = "1.3"

import md5, thread
from cPickle import PicklingError, dumps
from AccessControl import getSecurityManager
from OFS.Cache import Cache, CacheException
from time import time
from zLOG import LOG, DEBUG

ckey_args = "view_name, keywords, request"
ckey_code = """kw = ()
if keywords:
    kw = [(str(k), str(v)) for k, v in keywords.iteritems()]
    kw.sort()
return str(view_name), tuple(kw), \\
       str(request.other.get("AUTHENTICATED_USER"))
"""

class ProgrammerError(CacheException):
    """PEBKAC"""


class UncacheableDataError(CacheException):
    """Exception raised when uncacheable data is encountered"""


class MemoryCache(Cache):
    metadata_tax = 150
    def __init__(self, func_t, cache_size=0, max_ttl=0):
        self.ckey_func, self.security_context = func_t
        self.cache_size = cache_size
        self.max_ttl = max_ttl
        self.lock = thread.allocate_lock()
        self.data = {}     # d_key:[ref count, data]
        self.metadata = {} #   key:[d_key, size, (path), mtime, ctime, atime]
        self.size = 0      #        0      1     2       3      4      5

    def __delitem__(self, key):
        d_key, size = self.metadata[key][:2]
        del self.metadata[key]
        self.size -= self.metadata_tax
        if self.data[d_key][0] < 2:
            del self.data[d_key]
            self.size -= size
        else:
            self.data[d_key][0] -= 1

    def _digest(self, d, m=md5.new):
        try:
            p = dumps(d, 1)
        except PicklingError:
            raise UncacheableDataError("can't cache data that can't be pickled")
        return m(p).digest(), len(p)

    def _flush_LRU(self):
        if self.size > self.cache_size:
            keys = [(k, v[5]) for k, v in self.metadata.iteritems()]
            keys.sort(lambda a, b:  cmp(b[1], a[1]))
            keys = [k[0] for k in keys]
            while self.size > self.cache_size: del self[keys.pop()]

    def _make_key(self, o, path, view_name, keywords):
        f = self.ckey_func
        if f is None: raise ProgrammerError("cache key function has errors")
        sm = getSecurityManager()
        context = self.security_context
        sm.addContext(context)
        try:
            key_extras = f(view_name, keywords, o.REQUEST)
        finally:
            sm.removeContext(context)
        if __debug__: LOG("MemoryCache", DEBUG, "cache key", str(key_extras))
        return self._digest((path, key_extras))[0]

    def ZCache_invalidate(self, o):
        p = o.getPhysicalPath()
        d = len(p)
        self.lock.acquire()
        try:
            keys = [k for k, v in self.metadata.iteritems() if v[2][:d] == p]
            for k in keys: del self[k]
        finally:
            self.lock.release()

    def ZCache_get(self, o,
                   view_name='', keywords=None, mtime_func=None, default=None):
        key = self._make_key(o, o.getPhysicalPath(), view_name, keywords)
        stat = self.metadata.get(key)
        if stat is None: return default
        mtime = o.ZCacheable_getModTime(mtime_func)
        if stat[3] < mtime or self.max_ttl and stat[4] + self.max_ttl < time():
            self.lock.acquire()
            try:
                try: del self[key]
                except KeyError: pass # threadnapped
            finally:
                self.lock.release()
            return default
        stat[5] = time()
        return self.data.get(stat[0], (0, default))[1]

    def ZCache_set(self, o, data,
                   view_name='', keywords=None, mtime_func=None):
        d_key, size = self._digest(data)
        if size > self.cache_size: raise UncacheableDataError("data too large")
        now = time()
        path = o.getPhysicalPath()
        key = self._make_key(o, path, view_name, keywords)
        mtime = o.ZCacheable_getModTime(mtime_func)
        self.lock.acquire()
        try:
            exists = self.metadata.has_key(key)
            if exists and self.metadata[key][0] == d_key:
                self.metadata[key] = [d_key, size, path, mtime, now, now]
            else:
                if exists: del self[key]
                self.metadata[key] = [d_key, size, path, mtime, now, now]
                self.size += self.metadata_tax
                if self.data.has_key(d_key):
                    self.data[d_key][0] += 1
                else:
                    self.size += size
                    self.data[d_key] = [1, data]
                    self._flush_LRU()
        finally:
            self.lock.release()
