# MemoryCacheManager.py,v 1.3 2003/06/01 08:37:02 jamie v0.9
__version__ = "1.3"

import time, weakref
from AccessControl import ClassSecurityInfo, Permissions, ZopeGuards, \
                          full_write_guard, safe_builtins
from MemoryCache import MemoryCache, ckey_args, ckey_code
from OFS.Cache import CacheManager
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from RestrictedPython import PrintCollector, compile_restricted_function

import os
_www = os.path.join(os.path.dirname(__file__), "www")
_icon = os.path.join(_www, "memory.png")
del os
_product = __name__.split('.')[-2]
caches = {}
MANAGE_PERM = Permissions.view_management_screens
CHANGE_PERM = "Change cache managers"

class MemoryCacheManager(CacheManager, SimpleItem):
    """A MemoryCacheManager instance provides the UI to a MemoryCache."""
    security = ClassSecurityInfo()
    security.declareObjectPrivate()
    meta_type = "Memory Cache Manager"
    manage_options = (
        {"label":"Properties", "action":"manage_main",
         "help":(_product, "MemoryCacheManager.stx")},
        ) + CacheManager.manage_options + SimpleItem.manage_options

    def __init__(self, oid, title=''):
        self.__cid = "%s_%f" % (id(self), time.time())
        self._errors = self._warnings = ()
        self._settings = {"cache_size":262144, "max_ttl":0}
        self._source = ckey_code
        self._v_func = self._compile_func()
        self.id = oid
        self.title = title

    # Forced into inheriting from CopySupport thanks to the myriad of
    # assumptions Zope makes about objects in the zodb, this is here to
    # prevent users from trying to copy, move, or rename our instances.
    def _canCopy(self, op=0): return 0

    def _compile_func(self):
        c = compile_restricted_function(ckey_args, self._source or "pass",
                                        "cache_key_function", `self`)
        self._warnings = tuple(c[2])
        self._errors = tuple(c[1])
        if c[1]: return None
        l = {}
        exec c[0] in {
            "__debug__":__debug__,
            "__builtins__":safe_builtins,
            "_getattr_":ZopeGuards.guarded_getattr,
            "_getitem_":ZopeGuards.guarded_getitem,
            "_write_":full_write_guard,
            "_print_":PrintCollector,
            }, l
        return l.get("cache_key_function")

    def _func(self):
        if not hasattr(self, "_v_func"): self._v_func = self._compile_func()
        return self._v_func, self

    security.declarePrivate("ZCacheManager_getCache")
    def ZCacheManager_getCache(self):
        try:
            return weakref.proxy(caches[self.__cid])
        except KeyError:
            caches[self.__cid] = MemoryCache(self._func(), **self._settings)
            return weakref.proxy(caches[self.__cid])

    security.declarePrivate("manage_beforeDelete")
    def manage_beforeDelete(self, item, container):
        CacheManager.manage_beforeDelete(self, item, container)
        #cid = self.__cid # reduce the likelihood of triggering a race and
        #del self.__cid   # orphaning a MemoryCache in the caches dictionary?
        try:
            del caches[self.__cid]
        except KeyError:
            pass

    security.declareProtected(MANAGE_PERM, "manage_main")
    manage_main = PageTemplateFile("properties.zpt", _www)

    security.declareProtected(MANAGE_PERM, "getTitle","getSettings","getErrors",
                              "getWarnings","getSource","getStatus")
    def getTitle(self): return self.title
    def getSettings(self): return self._settings.copy()
    def getErrors(self): return self._errors
    def getWarnings(self): return self._warnings
    def getSource(self): return self._source
    def getStatus(self):
        try:
            c = caches[self.__cid]
        except KeyError:
            c = caches[self.__cid] = MemoryCache(self._func(), **self._settings)
        return {"size":c.size, "entries":len(c.metadata), "unique":len(c.data)}

    security.declareProtected(CHANGE_PERM, "manage_reinitialize")
    def manage_reinitialize(self, title, settings=None, REQUEST=None):
        """Reinitialize the cache with new settings."""
        if settings is None: settings = REQUEST.form
        source = str(settings["ckey_code"]).strip()
        if source: source += '\n'
        self._settings = {
            "cache_size":int(settings["cache_size"]),
            "max_ttl":int(settings["max_ttl"]),
            }
        if source != self._source:
            self._source = source
            self._v_func = self._compile_func()
        self.title = str(title)
        caches[self.__cid] = MemoryCache(self._func(), **self._settings)
        msg = "Cache reinitialized."
        return REQUEST is None and msg or \
               self.manage_main(self, REQUEST, manage_tabs_message=msg)


import Globals
Globals.InitializeClass(MemoryCacheManager)
del Globals

manage_addMemoryCacheManagerForm = PageTemplateFile("add.zpt", _www)
def manage_addMemoryCacheManager(self, id, REQUEST=None):
    """Add a MemoryCacheManager."""
    self._setObject(id, MemoryCacheManager(id,
        REQUEST is not None and str(REQUEST.form.get("title", '')) or ''))
    if REQUEST is not None: return self.manage_main(self, REQUEST)
